
import os, sys


# the reason for this is to modify the variable in place ... yuck
def get_paths(pathcfg, config_filename, data_base, new_rawmail, old_rawmail):
    # --------------------------------------------------------------
    # Organization locations
    pathcfg.config_file = config_filename
    pathcfg.bomail_data_base = data_base
    pathcfg.new_rawmail_dir = new_rawmail
    pathcfg.old_rawmail_dir = old_rawmail
    
    # save a copy of sent messages here if sending from localhost
    pathcfg.save_sent_dir = pathcfg.new_rawmail_dir
    
    # --------------------------------------------------------------
    # File locations: suggested not to edit
    
    # Locations of directories
    pathcfg.rel_email_dirname = "email/"
    pathcfg.rel_drafts_dirname = "drafts/"
    pathcfg.rel_attach_dirname = "attachments/"
    pathcfg.rel_trash_dirname = "trash/"
    pathcfg.rel_metadata_dirname = "metadata/"
    pathcfg.email_dir     = os.path.join(pathcfg.bomail_data_base, pathcfg.rel_email_dirname)
    pathcfg.attach_dir    = os.path.join(pathcfg.bomail_data_base, pathcfg.rel_attach_dirname)
    pathcfg.drafts_dir    = os.path.join(pathcfg.bomail_data_base, pathcfg.rel_drafts_dirname)
    pathcfg.trash_dir     = os.path.join(pathcfg.bomail_data_base, pathcfg.rel_trash_dirname)
    pathcfg.metadata_dir  = os.path.join(pathcfg.bomail_data_base, pathcfg.rel_metadata_dirname)
    
    # Locations of files that must be edited by hand
    pathcfg.handlers_file      = os.path.join(pathcfg.metadata_dir, "mail-handlers.txt")
    pathcfg.addr_alias_file    = os.path.join(pathcfg.metadata_dir, "aliases.txt")
    
    # Locations of files that may be edited by hand
    pathcfg.tags_file          = os.path.join(pathcfg.metadata_dir, "tags.txt")
    pathcfg.addr_book_file     = os.path.join(pathcfg.metadata_dir, "addr_book.txt")
    
    # Locations of files that should not be hand-edited
    pathcfg.openlist_file      = os.path.join(pathcfg.metadata_dir, "openlist.txt")
    pathcfg.scheduledlist_file = os.path.join(pathcfg.metadata_dir, "scheduledlist.txt")
    pathcfg.msg_ids_file       = os.path.join(pathcfg.metadata_dir, "msg_ids.txt")
    pathcfg.tab_config_file    = os.path.join(pathcfg.metadata_dir, "tab_config.py")
    
    # Locations of logs, these can be deleted whenever
    pathcfg.acts_log_file      = os.path.join(pathcfg.metadata_dir, "acts_log.txt")
    pathcfg.error_log_file     = os.path.join(pathcfg.metadata_dir, "error_log.txt")

#for d in [email_dir, attach_dir, drafts_dir, trash_dir, metadata_dir]:
#  if not os.path.exists(d):
#    os.makedirs(d)





####
# bomail.cli.send
#
# Send drafts.
####

import os
import sys
import mimetypes
import subprocess, email
import smtplib
import traceback
import socket

from email import encoders
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from bomail.config.config import pathcfg,sendcfg
import bomail.cli.mailfile as mailfile
import bomail.util.addr as addr
import bomail.util.util as util

usage_str = """
Reads filename(s) from stdin one per line and sends them.
send.py -h to print this help.
"""


# adapted from python library example: https://docs.python.org/3/library/email-examples.html
def add_attachments(msg, attach_list):
  for filename in attach_list:
    if not os.path.exists(filename):
      raise Exception("Did not send. Could not attach (does not exist): " + filename)
    if not os.path.isfile(filename):
      raise Exception("Did not send. Could not attach (is not a file): " + filename)
    ctype, encoding = mimetypes.guess_type(filename)
    if ctype is None or encoding is not None:
      ctype = 'application/octet-stream'
    maintype, subtype = ctype.split('/', 1)
    if maintype == 'text':
      with open(filename) as fp:
        sub_msg = MIMEText(fp.read(), _subtype=subtype)
    elif maintype == 'image':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEImage(fp.read(), _subtype=subtype)
    elif maintype == 'audio':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEAudio(fp.read(), _subtype=subtype)
    else:
      with open(filename, 'rb') as fp:
        sub_msg = MIMEBase(maintype, subtype)
        sub_msg.set_payload(fp.read())
      encoders.encode_base64(sub_msg)
    sub_msg.add_header('Content-Disposition', 'attachment', filename=os.path.split(filename)[1])
    msg.attach(sub_msg)


# convert file to a python message
# if multiple "from" addresses, use Sender header with sender_str
def file_to_msg(fname, sender_str, mgr):
  data = mgr.get_all(fname)
  attach_list = mgr.get_attachlist(fname)
  if len(attach_list) == 0:  # no attachments
    msg = MIMEText(data[mailfile.BODY_L], 'plain')
  else:
    msg = MIMEMultipart()
    msg.attach(MIMEText(data[mailfile.BODY_L], 'plain'))
    add_attachments(msg, attach_list)
  msg['Subject'] = data[mailfile.SUBJ_L]
  msg['From'] = data[mailfile.FROM_L]
  if len(addr.str_to_pairlist(data[mailfile.FROM_L])) > 1:
    msg['Sender'] = sender_str
  msg['To'] = data[mailfile.TO_L]
  msg['Date'] = email.utils.formatdate()
  msg['Message-ID'] = data[mailfile.MSG_ID_L]
  if len(data[mailfile.CC_L]) > 0:
    msg['CC'] = data[mailfile.CC_L]
  refs = data[mailfile.REFS_L]
  if len(data[mailfile.REPLY_L]) > 0:
    msg['Reply-To'] = data[mailfile.REPLY_L]
  if len(refs) > 0:
    msg['In-Reply-To'] = refs.split(", ")[-1]
    msg['References'] = refs

  recip_lists = data[mailfile.TO_L].split(", ")
  if len(data[mailfile.CC_L]) > 0:
    recip_lists += data[mailfile.CC_L].split(", ")
  if len(data[mailfile.BCC_L]) > 0:
    recip_lists += data[mailfile.BCC_L].split(", ")
  recip_addrs = [addr.str_to_pair(a)[1] for a in recip_lists]
  return recip_addrs, msg


# given list of relative filenames
# return results, msg
# where results is a list of success/not [True, False, ...]
def connect_and_send(user_ind, fnames, fname_to_senderstr, mgr):
  results = [False for f in fnames]

  try:
    if sendcfg.smtp_servernames[user_ind] == "localhost":
      serv = smtplib.SMTP("localhost")
    else:
      myhost = socket.gethostbyname(sendcfg.smtp_servernames[user_ind])  # prefer IPv4
      serv = smtplib.SMTP(myhost, sendcfg.smtp_ports[user_ind])
      serv.ehlo()
      serv.starttls()
      serv.ehlo()
      serv.login(sendcfg.smtp_usernames[user_ind], sendcfg.smtp_passwords[user_ind])
  except Exception as e:
    util.err_log("Sending email, error connecting to server.\n" + traceback.format_exc() + "\n")
    return results, "Error connecting to server: " + str(e)

  email_addr = sendcfg.email_addrs[user_ind]
  for i, fname in enumerate(fnames):
    recip_addrs, msg = file_to_msg(fname, fname_to_senderstr[fname], mgr)
    try:
      serv.sendmail(email_addr, recip_addrs, msg.as_string())
      if sendcfg.smtp_servernames[user_ind] == "localhost": # save the sent mail
        os.makedirs(pathcfg.new_rawmail_dir, exist_ok=True)
        try:
          rawname = util.mime_to_mailfile.get_msgid(m, fname)
          with open(os.path.join(pathcfg.new_rawmail_dir,rawname), "w") as rawfile:
            rawfile.write(msg.as_string())
        except:
          util.err_log("Error saving copy of sent mail:\n" + fname + "\n" + msg.as_string() + "\n")
      results[i] = True
    except Exception as e:
      serv.quit()
      util.err_log("Sending email, error while sending.\n" + traceback.format_exc() + "\n")
      return results, "Error sending message #" + str(i) + " [" + fname + "]: " + str(e)

  serv.quit()
  return results, ""


# given list of pairs, return who is the sender
# as an index into sendcfg.email_addrs
# along with the sender string it matched
def get_sender(fromlist):
  for pr in fromlist:
    try:
      user_ind = sendcfg.email_addrs.index(pr[1])
      return user_ind, addr.pair_to_str(pr[0], pr[1])
    except:
      pass
  raise Exception("no sender in list")


 

# given list of filenames
# return vector of successes [True, False, ...], err_msg
def main(flist, mgr):
  # sort them into different sender addresses and connect to each server once
  userind_to_fnames = [[] for a in sendcfg.email_addrs]
  fname_to_senderstr = {}
  for fname in flist:
    fromline = mgr.get(fname, mailfile.FROM_L)
    fromlist = addr.str_to_pairlist(fromline)
    try:
      user_ind, sender_str = get_sender(fromlist)
      userind_to_fnames[user_ind].append(fname)
      fname_to_senderstr[fname] = sender_str
    except:
      err_s = "Error! Cannot send from any of " + fromline + " in file " + fname
      util.err_log(err_s)
      return [], err_s    # sent no mails
    
  success = set()
  msg = ""
  for user_ind,fnames in enumerate(userind_to_fnames):
    res, msg = connect_and_send(user_ind, fnames, fname_to_senderstr, mgr)
    for i,good in enumerate(res):
      if good:
        success.add(fnames[i])
    if len(msg) > 0:
      break
  # return array with True if message sent
  return [f in success for f in flist], msg



def main_cli():
  if len(sys.argv) >= 2:
    print(usage_str)
    exit(0)

  flist = [f.strip() for f in sys.stdin.readlines()]
  res, err = main(flist, mailfile.MailMgr())
  print("Sent " + str(res) + "/" + str(len(flist)) + " messages.")
  if res < len(flist):
    print(err)


if __name__ == "__main__":
  main_cli()





####
# bomail.util.search_opts
#
# Parse search options!
####

import os
import shlex
import datetime
import re

import bomail.cli.mailfile as mailfile
import bomail.util.tags as tags
import bomail.util.addr as addr
import bomail.util.datestuff as datestuff
from bomail.util.listmail import ListOpts

options_str = """
All arguments are optional:
 -h               [print this help]
 -n 200           only return first 200 results
 -sortold         list by oldest (default is by newest)

 -a datestr       after date. datestr can be e.g. yyyy-mm-dd
 -b datestr       before date

 -open            state is open
 -scheduled       state is scheduled
 -closed          state is closed
 -draft           is a draft
 -sent            is sent from me
 -attach          has an attachment

 -notags          has no tags
 -to-me           addressed to any of my aliases in config file

Use quotes around "space separated strings" for these:
 -t str           has str as a tag
 -nt str          does not have str as a tag
 -to str          str is in to, cc, or bcc field
 -from str        str is in from field
 -subject str     str is in subject field
 str              str is in email somewhere

Can use -t "str1, str2" as shorthand for -t str1 -t str2.
"""


def with_quotes(q):
  if len(q) == 0:
    return q
  if q[0] == '"' and q[-1] == '"':
    return q  # assume it's properly quoted
  return '"' + q.replace('"', '\\"') + '"'

# add query q to search string
def get_new_search_str(old_str, q):
  return old_str + " " + with_quotes(q)


class SearchQuery:
  def __init__(self):
    self.max_num = -1  # indiciates unlimited
    self.reverse = False

    self.listopts = ListOpts()
    self.can_list_only = True      # True if ListOpts can list all matching files
    self.can_filename_only = True  # True if only need filename
    self.orig_after_str = None
    self.orig_before_str = None
    self.after_str = None
    self.before_str = None

    # Can be found by listing (without opening files)
    self.open = None
    self.scheduled = None
    self.draft = None
    self.attach = None

    # Cannot be found by listing
    self.sent = None
    self.closed = None

    self.tagset = None      # must match all
    self.not_tagset = None  # must not match any
    self.tolist = None      # must match all
    self.fromlist = None    # ditto
    self.subjectlist = None # ditto

    self.rough_regex_list = []  # regular expression to 'grep' with

    self.notags = None  # or True
    self.tome = None    # or True

    self.querylist = None  # must match all


  def parse(self, old_arglist):
    arglist = list(old_arglist)
    if "-n" in arglist:
      i = arglist.index("-n")
      del arglist[i]
      if len(arglist) > i:
        try:
          self.max_num = int(arglist[i])
        except:
          self.max_num = 0
        del arglist[i]

    if "-sortold" in arglist:
      self.reverse = True
      arglist.remove("-sortold")

    if "-a" in arglist:
      i = arglist.index("-a")
      del arglist[i]
      if len(arglist) > i:
        s = arglist[i]
        self.orig_after_str = s
        del arglist[i]
    if "-b" in arglist:
      i = arglist.index("-b")
      del arglist[i]
      if len(arglist) > i:
        s = arglist[i]
        self.orig_before_str = s
        del arglist[i]
  
    # check draft first because it comes from a separate list
    # so it is always handled by listmail, not filter
    if "-draft" in arglist:
      self.listopts.source = "draft"
      self.draft = True
      arglist.remove("-draft")
    if "-open" in arglist:
      self.can_filename_only = False
      if self.listopts.source == "all":
        self.listopts.source = "open"
      else:
        self.can_list_only = False  # must check both draft and open
      self.open = True
      arglist.remove("-open")
    if "-scheduled" in arglist:
      self.can_filename_only = False
      if self.listopts.source == "all":
        self.listopts.source = "scheduled"
      else:
        self.can_list_only = True
      self.scheduled = True
      arglist.remove("-scheduled")
    if "-attach" in arglist:
      self.can_filename_only = False
      if self.listopts.source == "all":
        self.listopts.source = "attach"
      else:
        self.can_list_only = False
      self.attach = True
      arglist.remove("-attach")

    if "-sent" in arglist:
      self.can_filename_only = False
      self.sent = True
      self.can_list_only = False
      arglist.remove("-sent")
    if "-closed" in arglist:
      self.can_filename_only = False
      self.closed = True
      self.can_list_only = False
      arglist.remove("-closed")

    while "-t" in arglist:
      self.can_filename_only = False
      self.can_list_only = False
      i = arglist.index("-t")
      del arglist[i]
      if len(arglist) < i:
        break
      tagstr = arglist[i]
      myset = tags.get_tagset_from_str(tagstr, include_folders=False)
      if len(myset) > 0:
        self.rough_regex_list.extend(myset)
        if self.tagset is None:
          self.tagset = myset
        else:
          self.tagset.update(myset)
      del arglist[i]

    while "-nt" in arglist:
      self.can_filename_only = False
      self.can_list_only = False
      i = arglist.index("-nt")
      del arglist[i]
      if len(arglist) < i:
        break
      tagstr = arglist[i]
      myset = tags.get_tagset_from_str(tagstr, include_folders=False)
      if len(myset) > 0:
        if self.not_tagset is None:
          self.not_tagset = myset
        else:
          self.not_tagset.update(myset)
      del arglist[i]

    while "-to" in arglist:
      self.can_filename_only = False
      self.can_list_only = False
      i = arglist.index("-to")
      del arglist[i]
      if len(arglist) < i:
        break
      mylist = [s.strip() for s in arglist[i].split(",")]
      self.rough_regex_list.extend(mylist)
      if self.tolist is None:
        self.tolist = mylist
      else:
        self.tolist.extend(mylist)
      del arglist[i]

    while "-from" in arglist:
      self.can_filename_only = False
      self.can_list_only = False
      i = arglist.index("-from")
      del arglist[i]
      if len(arglist) < i:
        break
      mylist = [s.strip() for s in arglist[i].split(",")]
      self.rough_regex_list.extend(mylist)
      if self.fromlist is None:
        self.fromlist = mylist
      else:
        self.fromlist.extend(mylist)
      del arglist[i]

    while "-subject" in arglist:
      self.can_filename_only = False
      self.can_list_only = False
      i = arglist.index("-subject")
      del arglist[i]
      if len(arglist) < i:
        break
      mylist = [s.strip() for s in arglist[i].split(",")]
      self.rough_regex_list.extend(mylist)
      if self.subjectlist is None:
        self.subjectlist = mylist
      else:
        self.subjectlist.extend(mylist)
      del arglist[i]

    if "-notags" in arglist:
      self.can_filename_only = False
      self.can_list_only = False
      self.notags = True
      arglist.remove("-notags")

    if "-to-me" in arglist:
      self.can_filename_only = False
      self.can_list_only = False
      self.tome = True
      arglist.remove("-to-me")

    if len(arglist) > 0:  # remaining strings are text search
      self.can_filename_only = False
      self.can_list_only = False
      self.rough_regex_list.extend(arglist)
      self.querylist = [re.compile(re.escape(s)) for s in arglist]

    for i in range(len(self.rough_regex_list)):
      self.rough_regex_list[i] = re.compile(re.escape(self.rough_regex_list[i]))


  # generate python code that only checks the files for items in the query
  def do_compile(self):
    slist = ["matched_inds = []\n"]
    slist.append("for ind,f in enumerate(flist):\n")

    # do filename-based stuff first
    # use that we store filenames with date in them (except drafts!)
    need_date = self.after_str is not None or self.before_str is not None
    if need_date:
      slist.append("  if not filename_matches_date(f, self.after_str, self.before_str):\n")
      slist.append("    continue\n")
    if self.draft is not None:
      slist.append("  if f[-5:] != \"draft\":\n")
      slist.append("    continue\n")

    # now do non-filename-based stuff
    if not self.can_filename_only:
      self.get_compile_data_checks(slist)

    # if we made it to the end of the loop, this one matches
    slist.append("  matched_inds.append(ind)\n")

    # stop the loop if we've found enough
    if self.max_num > 0:
      slist.append("  if len(matched_inds) >= " + str(self.max_num) + ":\n")
      slist.append("    break\n")


     # save the compiled search query to file
    with open("compile_log.txt", "w") as myf:
      myf.write("".join(slist))
    return compile("".join(slist), '<string>', 'exec')
    

  # for compiling a query, get all the checks that involve file data
  def get_compile_data_checks(self, slist):
    slist.append("  with open(f) as my_file:\n")
    slist.append("    contents = my_file.read()\n")

    # do a rough check through the file before exact checks
    if len(self.rough_regex_list) > 0:
    #### if self.case_sensitive...
      slist.append("  if not all([myre.search(contents) for myre in self.rough_regex_list]):\n")
      slist.append("    continue\n\n")

    slist.append("  data = mailfile.parse_file_contents(contents, f)\n")

    if self.notags:
      slist.append("  if len(data[mailfile.TAGS_L]) > 0:\n")
      slist.append("    continue\n")
    
    if self.open is not None:
      slist.append("  if not data[mailfile.STATE_L].startswith(\"open\"):\n")
      slist.append("    continue\n")
    if self.scheduled is not None:
      slist.append("  if not data[mailfile.STATE_L].startswith(\"scheduled\"):\n")
      slist.append("    continue\n")
    if self.closed is not None:
      slist.append("  if not data[mailfile.STATE_L].startswith(\"closed\"):\n")
      slist.append("    continue\n")
    if self.sent is not None:
      slist.append("  if not data[mailfile.SENT_L] == \"True\":\n")
      slist.append("    continue\n")
    if self.attach is not None:
      slist.append("  if len(data[mailfile.ATTACH_L]) == 0:\n")
      slist.append("    continue\n")

    if self.tome is not None:
      slist.append("  prlist_list = [addr.str_to_pairlist(data[j]) for j in [mailfile.TO_L, mailfile.CC_L, mailfile.BCC_L]]\n")
      slist.append("  if not any([any([addr.is_pair_me(pr) for pr in prlist]) for prlist in prlist_list]):\n")
      slist.append("    continue\n")

    # need to match all of a set
    if self.tagset is not None or self.not_tagset is not None:
      slist.append("  data_tagset = tags.get_tagset_from_str(data[mailfile.TAGS_L], include_folders=True)\n")
    if self.tagset is not None:
      slist.append("  if self.tagset.issubset(data_tagset):\n")
      slist.append("    continue\n")
    # if we match any of these, skip
    if self.not_tagset is not None:
      slist.append("  if self.not_tagset & data_tagset:\n")
      slist.append("    continue\n")
    # need to match all addresses
    if self.tolist is not None:
      slist.append("  if not all([s in data[mailfile.TO_L] or s in data[mailfile.CC_L] or s in data[mailfile.BCC_L] for s in self.tolist]):\n")
      slist.append("    continue\n")
    if self.fromlist is not None:
      slist.append("  if not all([s in data[mailfile.FROM_L] for s in self.fromlist]):\n")
      slist.append("    continue\n")
    if self.subjectlist is not None:
      slist.append("  if not all([s in data[mailfile.SUBJ_L] for s in self.subjectlist]):\n")
      slist.append("    continue\n")

    # every query must have all their strings somewhere
    if self.querylist is not None:
      #if self.case_sensitive:
        slist.append("  if not all([any([q.search(s, re.IGNORECASE) for s in data]) for q in self.querylist]):\n")
        slist.append("    continue\n")
      #else:
      #  slist.append("  if not all([any([re.search(q, s) for s in data]) for q in self.querylist])\n")
      #  slist.append("    continue\n")
  # end of extra search steps


  # convert relative dates and time zones to absolute UTC
  def interpret_dates(self):
    self.after_str = None
    self.before_str = None
    if self.listopts is not None:
      self.listopts.after_str = None
      self.listopts.before_str = None
    if self.orig_after_str is not None:
      self.after_str = datestuff.parse_schedstr_to_utcstr(self.orig_after_str)
      if self.listopts is not None:
        self.listopts.after_str = self.after_str[:10]
    if self.orig_before_str is not None:
      self.before_str = datestuff.parse_schedstr_to_utcstr(self.orig_before_str)
      if self.listopts is not None:
        self.listopts.before_str = self.before_str[:10]
                        

  # list all files matching basic part of query
  # (should filter them after this, unless self.can_list_only)
  def listmail(self, mgr):
    self.interpret_dates()
    flist = self.listopts.listmail(mgr)
    if self.reverse:
      flist.reverse()
    return flist

  # return list of indices matching our query on filenames
  def filter(self, mgr, flist):
    self.interpret_dates()
    compiled_obj = self.do_compile()
    namespace ={"re": re, "self": self, "addr": addr, "filename_matches_date": datestuff.filename_matches_date, "mailfile": mailfile, "tags": tags, "flist": flist}
    exec(compiled_obj, namespace)
    return namespace["matched_inds"]

  def search(self, mgr):
    result = self.listmail(mgr)
    if not self.can_list_only:
      matched_inds = self.filter(mgr, result)
      result = [result[i] for i in matched_inds] 
    if self.max_num > 0:
      return result[:self.max_num]
    return result





####
# bomail.guistuff.mygui
#
# The main UI object.
####

import curses
import os
import sys
import locale
import traceback

from bomail.config.config import pathcfg,guicfg

import bomail.util.search_opts as search_opts
import bomail.util.util as util

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.cli.process as process
import bomail.cli.check_sched as check_sched

from bomail.util.addr import AddrBook
from bomail.util.thread import ThreadMgr
from bomail.util.tags import TagMgr

import bomail.guistuff.display as display
import bomail.guistuff.get_txt as get_txt
import bomail.guistuff.gui_util as gui_util
from bomail.guistuff.acts import Acts
from bomail.guistuff.tabthread import TabThread
from bomail.guistuff.tabnothread import TabNoThread


default_tab_ind = 0
default_tab_searches = [
  "-open",
  "-draft",
  "-sent",
  ""]


class Gui:
  def __init__(self, screen):
    self.init_screen(screen)

    self.mail_mgr = mailfile.MailMgr()
    self.addr_book = AddrBook()
    if guicfg.threads_on:
      self.thread_mgr = ThreadMgr()
    else:
      self.thread_mgr = None
    self.tag_mgr = TagMgr()
    self.acts = Acts(self)

    self.unread_set = set()  # unread filenames

    try:
      with open(pathcfg.tab_config_file) as f:
        s = f.read()
        namespace = {}
        exec(s, namespace)
        self.tab_ind = namespace["tab_ind"]
        tab_searches = namespace["tab_searches"]
    except:
      self.tab_ind = default_tab_ind
      tab_searches = default_tab_searches
    if guicfg.threads_on:
      self.tabs = [TabThread(s, self) for s in tab_searches]
    else:
      self.tabs = [TabNoThread(s, self) for s in tab_searches]

    self.tabs[self.tab_ind].load()  # better load initial tab now, during loading screen
    self.curr_view = self.tabs[self.tab_ind]
    curses.flushinp()


  def init_screen(self, screen):
    self.screen = screen
    curses.curs_set(False)  # no cursor
    screen.scrollok(1)      # writing to last line/char is allowed
    curses.init_pair(guicfg.DEFAULT_CLR_PAIR, guicfg.foreground_color, guicfg.background_color)
    curses.init_pair(guicfg.BOMAIL_CLR_PAIR, guicfg.bomail_color, guicfg.background_color)
    curses.init_pair(guicfg.AUTHOR_CLR_PAIR, guicfg.author_color, guicfg.background_color)
    curses.init_pair(guicfg.TAGS_CLR_PAIR, guicfg.tags_color, guicfg.background_color)
    curses.init_pair(guicfg.THREAD_CLR_PAIR, guicfg.thread_color, guicfg.background_color)
      
    screen.keypad(True)
    
    self.attr = curses.color_pair(guicfg.BOMAIL_CLR_PAIR) | guicfg.bomail_attr
    self.screen.bkgd(" ", curses.color_pair(guicfg.BOMAIL_CLR_PAIR))
    self.tab_area = curses.newwin(curses.LINES - display.TOPLINES - display.BOTTOMLINES, curses.COLS, display.TOPLINES, 0)
    self.tab_area.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))
    self.note_area = curses.newwin(1, curses.COLS, curses.LINES - display.BOTTOMLINES + 1, 0)
    self.note_area.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))
    self.commands_area = curses.newwin(display.BOTTOMLINES-3, curses.COLS, curses.LINES - display.BOTTOMLINES + 3, 0)
    self.commands_area.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))

    display.draw_loading_screen(self)

  # after calling a program like 'less', reset curses stuff
  def reset_after_prog(self):
    curses.curs_set(False)
    curses.noecho()
    curses.cbreak()
    self.screen.keypad(True)

  def update_for_new(self, filelist):
    if len(filelist) == 0:
      return
    if guicfg.threads_on:
      prlist = [(self.mail_mgr.get(f, mailfile.MSG_ID_L),f) for f in filelist]
      self.thread_mgr.update_for_add(prlist, self.mail_mgr)
    for t in self.tabs:
      if t != self.curr_view:
        t.update_for_new(filelist)
    self.curr_view.update_for_new(filelist)

  def update_for_change(self, filelist):
    if len(filelist) == 0:
      return
    if guicfg.threads_on:
      prlist = [(self.mail_mgr.get(f, mailfile.MSG_ID_L), f) for f in filelist]
      self.thread_mgr.update_for_add(prlist, self.mail_mgr)
    for t in self.tabs:
      if t != self.curr_view:
        t.update_for_change(filelist)
    self.curr_view.update_for_change(filelist)

  # filelist are the now-deleted filenames
  # data_list are their data (note they are no longer in mail manager!)
  def update_for_trash(self, filelist, data_list):
    if len(filelist) == 0:
      return
    if guicfg.threads_on:
      self.thread_mgr.update_for_trash(filelist, data_list, self.mail_mgr)
    for t in self.tabs:
      if t != self.curr_view:
        t.update_for_trash(filelist)
    if self.curr_view.update_for_trash(filelist):  # if true, go back to tab
      self.change_tab_ind(self.tab_ind)

  def rewrite_tab_searches(self):
    with open(pathcfg.tab_config_file, "w") as f:
      f.write("tab_ind = " + str(self.tab_ind) + "\n\n")
      f.write("tab_searches = [\n")
      for t in self.tabs:
        f.write("  \"" + t.search_str.replace('"','\\"') + "\",\n")
      f.write("]\n")

  def change_tab_ind(self, new_ind):
    self.tab_ind = new_ind
    self.curr_view = self.tabs[self.tab_ind]
    if not self.curr_view.is_loaded:
      self.curr_view.load()

  def mark_read(self, filelist):
    for f in filelist:
      self.unread_set.discard(f)  # removes if present

  def mark_unread(self, filelist):
    self.unread_set.update(filelist)

  def go(self):
    keep_alive = True
    mode, note = "all", "Welcome"  # mode is all, view, or note
    while keep_alive:
      display.redraw(self, self.curr_view, mode, note)
      mode, note = "note", ""
      key = self.screen.getkey()
      keep_alive, mode, note = self.process_key(key)
    # done with program; save tab settings and quit
    self.rewrite_tab_searches()
    return self.tabs[self.tab_ind].get_curr_filenames()

  # given keypress
  # return 3 results:
  # 1) True if keep program alive, False to exit
  # 2) the mode to redraw
  # 3) the note to draw
  def process_key(self, key):
    if key == guicfg.QUIT_KEY:
      return False, "", ""

    # otherwise, return True, mode, note
    key_used, mode, note = self.process_toplevel_key(key)
    if key_used:
      return True, mode, note

    curr_filenames = self.curr_view.get_curr_filenames()
    key_used, mode, note = self.process_fileacts_key(key, curr_filenames)
    if key_used:
        return True, mode, note
    # else
    self.curr_view, mode, note = self.curr_view.process_key(key)
    return True, mode, note


  # return (True if we use the keypress, False o.w.), mode, note
  def process_toplevel_key(self, key):
    ## Navigating tabs
    if key >= "0" and key <= "9":
      key_num = int(key)
      new_ind = key_num - 1 if key_num > 0 else 9   # key "0" goes to the 10th tab
      if new_ind < len(self.tabs):
        self.change_tab_ind(new_ind)
        return True, "all", ""
      # else
      return True, "note", "Can't switch: tab " + str(key_num) + " nonexistent"
    elif key == guicfg.TAB_LEFT_KEY:
      if self.tab_ind > 0:
        self.change_tab_ind(self.tab_ind - 1)
        return True, "all", ""
      # else
      return True, "note", "Already at leftmost tab."
    elif key == guicfg.TAB_RIGHT_KEY:
      if self.tab_ind < len(self.tabs) - 1:
        self.change_tab_ind(self.tab_ind + 1)
        return True, "all", ""
      # else
      return True, "note", "Already at rightmost tab."

    ## Adding, removing, shifting tabs
    elif key == guicfg.ADD_TAB_KEY:
      new_str = get_txt.get_tab_search_txt(self, "")
      if new_str is None:
        return True, "all", "Cancelled"
      # else
      mode, note = self.acts.do(("add tab", self.tab_ind + 1, new_str))
      return True, mode, note
    elif key == guicfg.REMOVE_TAB_KEY:
      if len(self.tabs) > 0:
        mode, note = self.acts.do(("remove tab", self.tab_ind, self.tabs[self.tab_ind]))
        return True, mode, note
      # else
      return True, "note", "Could not remove the only tab."
    elif key == guicfg.MOVE_TAB_LEFT_KEY:
      if self.tab_ind > 0:
        mode, note = self.acts.do(("move tab", self.tab_ind, -1))
        return True, mode, note
      # else
      return True, "note", "Tab is already leftmost."
    elif key == guicfg.MOVE_TAB_RIGHT_KEY:
      if self.tab_ind < len(self.tabs) - 1:
        mode, note = self.acts.do(("move tab", self.tab_ind, 1))
        return True, mode, note
      # else
      return True, "note", "Tab is already rightmost."
    elif key == guicfg.EDIT_TAB_KEY:
      old_str = self.tabs[self.tab_ind].search_str
      new_str = get_txt.get_tab_search_txt(self, old_str)
      if new_str is None:
        return True, "all", "Cancelled"
      # else
      mode, note = self.acts.do(("edit tab", self.tab_ind, new_str, old_str))
      return True, mode, note
    elif key == guicfg.SEARCH_IN_TAB_KEY:
      old_str = self.tabs[self.tab_ind].search_str
      q = get_txt.get_search_in_tab_txt(self)
      if q is None or len(q) == 0:
        return True, "all", "Cancelled"
      # else
      new_str = search_opts.get_new_search_str(old_str, q)
      mode, note = self.acts.do(("add tab", self.tab_ind + 1, new_str))
      return True, mode, note

    ## Undo, redo
    elif key == guicfg.UNDO_KEY:
      if len(self.acts.acts) > 0 and self.acts.act_ind >= 0:
        mode, note = self.acts.undo()
        return True, mode, note
      # else
      return True, "note", "Nothing to undo."
    elif key == guicfg.REDO_KEY:
      if len(self.acts.acts) > 0 and self.acts.act_ind + 1 < len(self.acts.acts):
        mode, note = self.acts.redo()
        return True, mode, note
      # else
      return True, "note", "Nothing to redo."

    ## Get updates
    elif key == guicfg.GET_KEY:
      display.redraw_note(self, "Getting new mail and updates (please wait)...")
      new_files = process.main(self.mail_mgr, self.tag_mgr)
      unsched_files = check_sched.main(self.mail_mgr)
      self.update_for_new(new_files)
      self.update_for_change(unsched_files)
      self.mark_unread(new_files)
      self.mark_unread(unsched_files)
      curses.flushinp()  # it might've been a long wait, flush keypresses
      unsched_str = "" if len(unsched_files) == 0 else ", de-scheduled " + str(len(unsched_files)) + " emails"
      num_sent = sum([1 if self.mail_mgr.get(f,mailfile.SENT_L)=="True" else 0 for f in new_files])
      sent_str = "" if num_sent == 0 else " (" + str(num_sent) + " sent)"
      return True, "all", "Got " + str(len(new_files)) + " new emails" + sent_str + unsched_str

    ## we didn't use the keypress
    return False, "", ""


  # return (True if the key was used, False o.w.), mode, note
  def process_fileacts_key(self, key, curr_filelist):
    mode, note = "", ""

    ## If no files, only valid operation is to write a new blank draft
    if len(curr_filelist) == 0:
      if key == guicfg.WRITE_KEY:
        mode, note = gui_util.go_compose_draft(self, "n", None)
        return True, mode, note
      return True, "note", "Key not recognized / nothing to do"

    ## Changing state
    if key == guicfg.OPEN_KEY:
      mode, note = self.acts.do(("open", curr_filelist, [self.mail_mgr.get(f, mailfile.STATE_L) for f in curr_filelist]))
    elif key == guicfg.CLOSE_KEY:
      mode, note = self.acts.do(("closed", curr_filelist, [self.mail_mgr.get(f, mailfile.STATE_L) for f in curr_filelist]))
    elif key == guicfg.SCHEDULE_KEY:
      mode, note = gui_util.go_schedule(self, curr_filelist)
    elif key == guicfg.TRASH_KEY:
      mode, note = self.acts.do(("trash", curr_filelist, [self.mail_mgr.get_all(f) for f in curr_filelist]))

    ## Tags
    elif key == guicfg.ADD_TAGS_KEY:
      mode, note = gui_util.go_add_tags(self, curr_filelist)
    elif key == guicfg.REMOVE_TAGS_KEY:
      mode, note = gui_util.go_remove_tags(self, curr_filelist)
    
    ## Writing and sending
    elif key == guicfg.WRITE_KEY:
      filename = curr_filelist[0]
      mode, note = gui_util.go_write_key(self, filename)
    elif key == guicfg.SEND_KEY:
      if all([filename[-5:] == "draft" for filename in curr_filelist]):
        mode, note = gui_util.go_send(self, curr_filelist)
      else:
        if len(curr_filelist) == 1:
          mode, note = "note", "Cannot send: not a draft"
        else:
          mode, note = "note", "Cannot send: not all drafts"

    # we used the key if we set the mode
    return (len(mode) > 0), mode, note


def main(screen):
  try:
    filenames = Gui(screen).go()
    output = "\n".join(filenames)
  except:
    output = traceback.format_exc()
  return output




####
# bomail.config.sendconfig
#
# Main configuration file.
# When loaded, it tries to open config file and read it; if unsuccessful,
# it runs a short setup program asking questions.
####

import os, sys

from bomail.config.conf_setup import options_dict


# --------------------------------------------------------------
# User info

# return True on success, False on failure
# for user 1, there's no suffix e.g. name = MY_NAME
# for users 2 on, there are, e.g. name2 = MY_OTHER_NAME
def get_user_info(sendcfg, options_dict, error_log_file, j):
  suffix = "" if j<=1 else str(j)
  if not ("name" + suffix) in options_dict:
    return False
  name = options_dict["name" + suffix]
  email_addr = options_dict["email_addr" + suffix]
  servname = options_dict["smtp_servername" + suffix]
  if servname == "localhost":
    # no need to get password, port, etc
    sendcfg.smtp_servernames.append(servname)
    sendcfg.names.append(name)
    sendcfg.email_addrs.append(email_addr)
    return True
  if ("smtp_userpass_file" + suffix) in options_dict:
    try:
      with open(options_dict["smtp_userpass_file" + suffix]) as f:
        for line in f:
          if line.startswith("username ="):
            uname = line[line.index("=")+1:].strip()
          if line.startswith("password ="):
            pword = line[line.index("=")+1:].strip()
    except:
      try:
        with open(error_log_file, "a") as f:
          f.write("\n================\n")
          f.write("---- " + datestuff.get_local_nowstr() + ":\n")
          f.write("config: Could not read username/password from " + options_dict["smtp_userpass_file" + suffix])
          f.write("\n\n")
      except:
        pass
      return False
  else: # get username and password straight from the config
    uname = options_dict["smtp_username" + suffix]
    pword = options_dict["smtp_password" + suffix]

  sendcfg.names.append(name)
  sendcfg.email_addrs.append(email_addr)
  sendcfg.smtp_servernames.append(servname)
  sendcfg.smtp_ports.append(int(options_dict["smtp_port" + suffix]))
  sendcfg.smtp_usernames.append(uname)
  sendcfg.smtp_passwords.append(pword)
  return True

def get_sendinfo(sendcfg, options_dict, error_log_file):
  sendcfg.names = []
  sendcfg.email_addrs = []
  sendcfg.smtp_servernames = []
  sendcfg.smtp_ports = []
  sendcfg.smtp_usernames = []
  sendcfg.smtp_passwords = []

  # first get myname, then myname2, myname3, ...
  get_user_info(sendcfg, options_dict, error_log_file, 0)
  j = 2
  while get_user_info(sendcfg, options_dict, error_log_file, j):
    j += 1

  sendcfg.my_aliases = []
  if "alias_addresses" in options_dict:
    sendcfg.my_aliases = [s.strip() for s in options_dict["alias_addresses"].split(",")]





####
# bomail.util.addr
#
# Utilities for dealing with email addresses
# and the "address book", which is stored in config.addr_book_file.
####

import sys
import os
from bisect import bisect_left

from bomail.config.config import pathcfg,sendcfg
import bomail.cli.mailfile as mailfile
import bomail.util.merge_lines as merge_lines
import bomail.util.util as util

####
# Address book file format: each line is
#    email_addr, N M
# where N and M are integers for num emails sent and received respectively.
# i.e. the user has sent N emails to them and received M emails from them.
#
# Python 'types' used:
# pr      A pair of strings of the form (display name, local@domain)
# prlist  A list of the form [pr, pr, ...]
#
# Plain text stuff:
#
# email_addr  An email address string of the form
#             display name <local@domain>
#             or local@domain
#             (see notes on RFC 5322 below)
####

# Ideally: email addresses could be any RFC 5322 valid recipient
# Currently: groups are not supported
# Currently: comments and quoted parts of email addresses not supported
#   (if you haven't read RFC 5322, you don't know what that is and that's ok)
# So for us, email address is local@domain where
# local and domain can contain any characters except @ < > and space

# An address-list is a comma-separated list of address strings.
# An address string is either
# (1) [display-name] <addr-spec>
# (2) addr-spec
# and addr-spec is a valid email address, which here we parse as above
# A display name can be a space-separated list of words if it doesn't contain special characters,
# else it must be a quoted string.

# My parsing rules:
# if there is no space character before the first @, then it's an addr-spec (possibly with < > around it)
# otherwise, find everything up to that space character, strip whitespace; that's the display name

RFC5322_specials = ["(" , ")" , 
                    "<" , ">",
                    "[" , "]",
                    ":" , ";",
                    "@" , "\\",
                    "," , ".",
                    '"']

RFC5322_specials_set = set(RFC5322_specials)

WHITESPACE = [' ', '\t', '\n', '\r']
MYSKIP_CHARS = [' ', '\t', '\n', '\r', ',']

def is_valid_quotestr(s):
  if len(s) < 2:
    return False
  if s[0] != '"':
    return False
  i = 1
  while i < len(s)-1:
    if s[i] == '\\':
      i += 1  # skip escaped character
    elif s[i] == '"':
      return False  # error: ended early 
    i += 1
  if i != len(s)-1 or s[-1] != '"':
    return False
  return True

def contains_specials(s):
  for c in s:
    if c in RFC5322_specials_set:
      return True
  return False

# check it either contains no special characters, or is already quoted properly
def needs_quoting(s):
  if not contains_specials(s):
    return False  # it's fine
  if is_valid_quotestr(s):
    return False
  return True

# return s, beginning and ending with a quote, and with any \ or " chars quoted
def quote(s):
  newlist = ['"']
  for c in s:
    if c == '\\' or c == '"':
      newlist.append('\\')
    newlist.append(c)
  newlist.append('"')
  return ''.join(newlist)

# This is just to guess when the user is the recipient of a message (err on the side of yes)
# canonical version that ignores case, dots, and +stuff in
# email address
# input is the second part of a pair (e.g. "local@domain")
def canon(s):
  if "@" not in s:
    return s  # no idea what's going on
  words = s.split("@")
  local, domain = words[0], words[1]
  local = local.lower()
  if "+" not in local and "." not in local:
    return s
  if "+" in local:
    local = local[:local.index("+")]
  if "." in local:
    local = local.replace(".","")
  return local + "@" + domain


# GLOBAL VARIABLE
# the user's list of email addresses
my_canon_addr_list = [canon(a) for a in sendcfg.my_aliases + sendcfg.email_addrs]



# should work on <local@domain> or local@domain
# return address (with < > stripped), end_index
# return None, None if could not parse
def get_first_addr_from_nodisp(addr_str):
  if "@" not in addr_str:
    return None, None  # no idea what to do
  at_ind = addr_str.index("@")
  end_ind = at_ind + 1
  while end_ind < len(addr_str) and addr_str[end_ind] not in [' ', ',']:
    end_ind += 1
  mypart = addr_str[:end_ind].strip()
  if len(mypart) >= 1 and mypart[0] == "<":
    mypart = mypart[1:]
  if len(mypart) >= 1 and mypart[-1] == ">":
    mypart = mypart[:-1]
  return mypart, end_ind

# return pair, end_index
# assume first character is "
# return None, None on fail
def get_first_pr_from_quoted(addr_str):
  if len(addr_str) <= 0 or addr_str[0] != '"':
    return None, None
  i = 1
  while i < len(addr_str) and addr_str[i] != '"':
    if addr_str[i] == "\\":
      i += 2
    else:
      i += 1
  if i >= len(addr_str):
    return None, None  # no address part!
  disp_name = addr_str[1:i].strip()
  addr_start = i+1
  addr_part, an_end_ind = get_first_addr_from_nodisp(addr_str[addr_start:])
  if addr_part is None:
    return None, None
  return [disp_name, addr_part], addr_start + an_end_ind

# return pair, end_index
# return None, None if fail
def get_first_pr_from_nonquoted(addr_str):
  if "@" not in addr_str:
    return None, None
  addr_start = 0
  if " " not in addr_str:
    disp_name = ''
  else:
    at_ind = addr_str.index("@")
    split_ind = addr_str.rfind(" ", 0, at_ind)
    disp_name = addr_str[:split_ind].strip()
    addr_start = split_ind + 1
  addr_part, an_end_ind = get_first_addr_from_nodisp(addr_str[addr_start:])
  if addr_part is None:
    return None, None
  return [disp_name, addr_part], addr_start + an_end_ind
  

# given plain-text comma-separated list of email addresses,
# return list of pairs
def str_to_pairlist(addr_str):
  addr_str = addr_str.strip()
  prlist = []
  index = 0
  while index < len(addr_str):
    while index < len(addr_str) and addr_str[index] in MYSKIP_CHARS:
      index += 1
    if index >= len(addr_str):
      break
    at_ind = addr_str.find('@', index)
    if at_ind < 0:
      break  # weird, rest of string has no @ sign
    elif addr_str[index] == '"':
      pair, mylen = get_first_pr_from_quoted(addr_str[index:])
    else:
      space_ind = addr_str.find(' ', index)
      if space_ind < 0 or at_ind < space_ind:  # no space in this email
        addr_part, mylen = get_first_addr_from_nodisp(addr_str[index:])
        pair = None if addr_part is None else ["", addr_part]
      else:
        pair, mylen = get_first_pr_from_nonquoted(addr_str[index:])
    if pair is None:
      break  # couldn't parse this last one
    elif needs_quoting(pair[0]):
      pair[0] = quote(pair[0])
    prlist.append((pair[0], pair[1]))
    index += mylen
  return prlist



#def str_to_pairlist2(addr_str, prlist=None):
#  if prlist is None:
#    prlist = []
#  i = 0
#  while i < len(addr_str) and addr_str[i] in MYSKIP_CHARS:
#    i += 1
#  addr_str = addr_str[i:].strip()
#  if '@' not in addr_str:
#    return prlist  # no idea what this is
#  elif addr_str[0] == '"':
#    pair, end_ind = get_first_pr_from_quoted(addr_str)
#  elif ' ' not in addr_str or addr_str.index("@") < addr_str.index(" "):
#    addr_part, end_ind = get_first_addr_from_nodisp(addr_str)
#    pair = None if addr_part is None else ["", addr_part]
#  else:
#    pair, end_ind = get_first_pr_from_nonquoted(addr_str)
#  if pair is None:
#    return prlist  # couldn't parse the last one
#
#  if needs_quoting(pair[0]):
#    pair[0] = quote(pair[0])
#  prlist.append((pair[0], pair[1]))
#  if end_ind >= len(addr_str):
#    return prlist  # done
#  return str_to_pairlist(addr_str[end_ind:], prlist)


# opposite direction
def pairlist_to_str(pair_list):
  return ", ".join([pair_to_str(*r) for r in pair_list])

def str_to_pair(auth_str):
  prlist = str_to_pairlist(auth_str)
  if len(prlist) > 0:
    return prlist[0]
  util.err_log("Could not get pair from: " + str(auth_str))
  return ('','')

def pair_to_str(auth, addr):
  if auth == "":
    return addr
  else:
    return auth + " <" + addr + ">"

# given a line from address book,
# return email pair and N, M
def str_to_quad(s):
  if "," not in s:
    p = str_to_pair(s)
    return (p[0], p[1], 0, 0)
  l = s.split(",")
  p = str_to_pair(l[0].strip())
  try:
    counts = l[1].split()
    return (p[0], p[1], int(counts[0].strip()), int(counts[1].strip()))
  except:
    return (p[0], p[1], 0, 0)


# given email pair and two counts,
# return a line for the address book
def quad_to_str(q):
  return pair_to_str(q[0], q[1]) + " , " + str(q[2]) + " " + str(q[3])


# check if the emails go to the same place
# not used! deprecated
#def is_recip_eq(p1, p2):
#  return canon(p1) == canon(p2)

def is_pair_me(pr):
  return canon(pr[1]) in my_canon_addr_list

def is_str_me(s):
  return is_pair_me(str_to_pair(s))


# pair_to_count is a dictionary mapping (name, email) to (#sent, #recv)
def write_addr_file(pair_to_count):
  # add my entry_quads to the dict, then flatten it to a list and write to disk
  entry_quads = [(p[0], p[1], c[0], c[1]) for p,c in pair_to_count.items()]
  entry_quads.sort()
  s = "\n".join([quad_to_str(q) for q in entry_quads])
  with open(pathcfg.addr_book_file, "w") as f:
    f.write(s)



# Mostly, this class is just used to read the list of addresses
# and to write new ones into the text file
class AddrBook:
  def __init__(self):
    self.pair_to_count = {}    # map (name, email_addr) to (send_count, recv_count)
    self.address_to_pair = {}    # map email_addr to (name, email_addr)
    self.load()


  def load(self):
    if not os.path.exists(pathcfg.addr_book_file):
      with open(pathcfg.addr_book_file, "w") as f:
        pass
    else:
      with open(pathcfg.addr_book_file) as f:
        lines = f.readlines()
      for l in [ll for ll in lines if ll.strip() != ""]:
        q = str_to_quad(l)
        self.pair_to_count[(q[0], q[1])] = (q[2], q[3])
        self.address_to_pair[q[1]] = (q[0], q[1])


  def check_for_new(self, prlist):
    added = False
    for pr in prlist:
      tuple_pr = (pr[0], pr[1])
      if tuple_pr not in self.pair_to_count:
        self.pair_to_count[tuple_pr] = (0, 0)
        self.address_to_pair[pr[1]] = tuple_pr
        added = True
    if added:
      self.rewrite_file()


  # delete all addresses in pairlist
  def remove_addresses(self, pairlist):
    rmlist = [p for p in pairlist if p in self.pair_to_count]
    if len(rmlist) != 0:
      for p in rmlist:
        del self.pair_to_count[p]
        if p[1] in self.address_to_pair:
          del self.address_to_pair[p[1]]
      self.rewrite_file()


  # delete all addresses in pairlist and add their
  # send/receive counts to merge_to
  def merge_addresses(self, merge_to, pairlist):
    c1, c2 = self.pair_to_count[merge_to]
    for pr in pairlist:
      t1, t2 = self.pair_to_count[pr]
      c1 += t1
      c2 += t2
      del self.pair_to_count[pr]
      del self.address_to_pair[pr[1]]
    self.pair_to_count[merge_to] = (c1, c2)
    self.rewrite_file()


  # look up pr in the book
  # if not present, but pr only contains an address
  # that is present, then change to that address
  def lookup_and_change(self, pr):
    if pr in self.pair_to_count:
      return pr
    if pr[0] == "" and pr[1] in self.address_to_pair:
      return self.address_to_pair[pr[1]]
    return pr


  # update our data to include any new addresses and new
  # send/receive counts
  def update_for_new(self, maildatas):
    for data in maildatas:
      fromlist = str_to_pairlist(data[mailfile.FROM_L])
      tolist = str_to_pairlist(data[mailfile.TO_L]) + str_to_pairlist(data[mailfile.CC_L]) + str_to_pairlist(data[mailfile.BCC_L])
      fromlist = [self.lookup_and_change(pr) for pr in fromlist]
      tolist = [self.lookup_and_change(pr) for pr in tolist]
      # make default entries for everything we haven't seen
      for p in [pr for pr in fromlist+tolist if pr not in self.pair_to_count]:
        self.pair_to_count[p] = (0, 0)
        self.address_to_pair[p[1]] = p
      # increment counts depending on if sent or received
      if data[mailfile.SENT_L] == "True":
        for p in tolist:
          cnt = self.pair_to_count[p] if p in self.pair_to_count else (0, 0)
          self.pair_to_count[p] = (cnt[0] + 1, cnt[1])
      else:
        for p in fromlist:
          cnt = self.pair_to_count[p] if p in self.pair_to_count else (0, 0)
          self.pair_to_count[p] = (cnt[0], cnt[1] + 1)
    self.rewrite_file()

 
  # save our address book to the text file
  def rewrite_file(self):
    write_addr_file(self.pair_to_count)





# change the configuration file automatically

import sys, os
import tempfile

import bomail.config.config as config


# change a line of the bomail.conf file of the form "opt_name = value"
def change_option(opt_name, new_value):
  with open(config.config_file) as fromf:
    lines = fromf.readlines()
  for i,line in enumerate(lines):
    if line.startswith(opt_name + " ="):
      lines[i] = opt_name + " = " + new_value + "\n"  # need newline
  with open(config.config_file, "w") as tof:
    tof.write("".join(lines))





####
# bomail.util.listmail
#
# (Quickly) list emails matching date range or simple state queries.
# The idea is it's fast because it doesn't have to open the files.
####

import sys
import os.path

from bomail.config.config import pathcfg
import bomail.util.datestuff as datestuff
import bomail.cli.mailfile as mailfile


# after and before may be none
# returns absolute file names, I guess
def list_all(after, before, mgr, dirname=None):
  if dirname is None:
    dirname = pathcfg.email_dir
  if not os.path.exists(dirname):
    return []
  all_years = os.listdir(dirname)
  if len(all_years) == 0:
    return []
  output = []
  for year_str in all_years:
    if after is not None and year_str < after[:4]:   continue
    if before is not None and year_str > before[:4]: continue
    year_prefix = year_str + "-"
    year_dir = os.path.join(dirname, year_str)
    all_days = os.listdir(year_dir)
    for day_str in all_days:
      date_str = year_prefix + day_str
      if after is not None and date_str[:len(after)] < after:   continue
      if before is not None and date_str[:len(before)] > before: continue
      day_dir = os.path.join(year_dir, day_str)
      for s in os.listdir(day_dir):
        output.append(os.path.join(day_dir, s))
  return output

# openlist is sorted by reverse lexicographic order
# return list of relative filenames
def list_open(after, before, mgr):
  if not os.path.exists(pathcfg.openlist_file):
    return []
  output = []
  with open(pathcfg.openlist_file) as f:
    for line in f:
      fname = line.strip()
      if datestuff.datestr_matches(datestuff.get_date_from_filename(fname), after, before):
        output.append(fname)
  return output

# scheduledlist is sorted by schedule date, not file date!
# have to check date of file!
# lines have the form "scheduledate emailfilename"
def list_scheduled(after, before, mgr):
  if not os.path.exists(pathcfg.scheduledlist_file):
    return []
  output = []
  with open(pathcfg.scheduledlist_file) as f:
    for fullline in f:
      line = fullline.strip()
      if len(line) == 0 or " " not in line:
        continue
      ind = line.index(" ")
      fname = line[ind+1:].strip()
      ds = datestuff.get_date_from_filename(fname)
      if datestuff.datestr_matches(ds, after, before):
        output.append(fname)
  return output

def list_draft(after, before, mgr):
  return list_all(after, before, mgr, pathcfg.drafts_dir)

# similar to list_all, but use the attach directories (only the ones that exist)
def list_attachonly(after, before, mgr):
  if not os.path.exists(pathcfg.attach_dir):
    return []
  all_years = os.listdir(pathcfg.attach_dir)
  if len(all_years) == 0:
    return []
  output = []
  for year_str in all_years:
    if after is not None and year_str < after[:4]:   continue
    if before is not None and year_str > before[:4]: continue
    year_prefix = year_str + "-"
    year_dir = pathcfg.attach_dir + year_str + "/"
    all_days = os.listdir(year_dir)
    for day_str in all_days:
      date_str = year_prefix + day_str
      if after is not None and date_str[:len(after)] < after:   continue
      if before is not None and date_str[:len(before)] > before: continue
      day_dir = year_dir + day_str + "/" 
      for s in os.listdir(day_dir):
        if s[-1] == "/":
          s = s[:-1]
        output.append(pathcfg.email_dir + year_str + "/" + day_str + "/" + s + mailfile.MAIL_EXT)
        allattach = os.listdir(os.path.join(day_dir,s))
  return output


class ListOpts:
  def __init__(self):
    self.after_str = None
    self.before_str = None
    self.source = "all"

  def listmail(self, mgr):
    if self.source == "all":
      return list_all(self.after_str, self.before_str, mgr)
    elif self.source == "open":
      return list_open(self.after_str, self.before_str, mgr)
    elif self.source == "scheduled":
      return list_scheduled(self.after_str, self.before_str, mgr)
    elif self.source == "draft":
      return list_draft(self.after_str, self.before_str, mgr)
    elif self.source == "attach":
      return list_attachonly(self.after_str, self.before_str, mgr)

  



####
# bomail.cli.meta
#
# A little utility to migrate or re-read metadata.
####

import os,sys,shutil,traceback

import bomail.config.config as config
from bomail.config.config import pathcfg

import bomail.config.edit_config as edit_config

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.cli.chstate as chstate

import bomail.util.addr as addr
import bomail.util.msgids as msgids
from bomail.util.tags import TagMgr

usage_str = """
To migrate bomail's data folder to a different path:
This moves to path/bomail/email, path/bomail/metadata, etc.

  bomail meta -m path/

To delete the log files (l as in log):

  bomail meta -l

To update some kinds of metadata (multiple options may be included):
  bomail meta -h         show this help and exit
  bomail meta -b         re-generate address book (addresses and counts)
  bomail meta -t         re-generate list of tags
  bomail meta -d         update all thread (conversation) relationships
  bomail meta -i         re-generate message IDs
  bomail meta -s         re-generate open list and scheduled list

  bomail meta -a         do all of the above

Why you would want to update metadata:
 -if a tag is obsolete, bomail meta -t will remove it from suggestions
  (can also edit metadata/tags.txt)
 -running bomail meta -d may help bomail find thread relationships,
  it tries to update dynamically so this probably won't help.
 -the others should not be necessary unless there was a crash/error.
"""

# maximum messages to hold in memory before releasing them
MAX_MSGS_IN_MEM = 10000


# update thread stuff
def update_refby_info(ids_to_fnames, ids_to_refbysets):
  mgr = mailfile.MailMgr()
  msgs_in_mem = 0
  for msgid,refby in ids_to_refbysets.items():
    if msgid not in ids_to_fnames:  # reference a message we don't have
      continue
    mgr.set(ids_to_fnames[msgid], mailfile.REF_BY_L, ", ".join(sorted(list(refby))))
    msgs_in_mem += 1
    if msgs_in_mem >= MAX_MSGS_IN_MEM:
      mgr = mailfile.MailMgr()
      msgs_in_mem = 0

# assume both oldpre and newpre are directories and s is a path
def replace_prefix(s, oldpre, newpre, helpstr):
  if not s.startswith(oldpre):
    # problem...
    return s
  else:
    return os.path.join(newpre,s[len(oldpre):])


# single pass through all of our files
# newdir is the new data base of bomail
def update(do_tags, do_book, do_threads, do_ids, do_lists, olddir=None, newdir=None):
  mgr = mailfile.MailMgr()
  all_fnames = search.search_argstr("", mgr)
  used_tagset = set()
  addrpr_to_counts = dict()  # map (name, email) to (send_count, recv_count)
  ids_to_fnames = dict()
  ids_to_refbysets = dict()  # map id to all id's it is referenced by
  openlist = []  # filename
  scheduledlist = [] # pair (filename, datestr)
  msgs_in_mem = 0
  for fname in all_fnames:
    if do_tags:
      my_tagset = mgr.get_tagset(fname)
      used_tagset.update(my_tagset)

    if do_book:
      if mgr.get(fname, mailfile.SENT_L) == "True":
        for line in [mailfile.TO_L,mailfile.CC_L,mailfile.BCC_L]:
          prs = addr.str_to_pairlist(mgr.get(fname, line))
          for rec in prs:
            if rec in addrpr_to_counts:
              addrpr_to_counts[rec][0] += 1
            else:
              addrpr_to_counts[rec] = [1,0]
      else:  # we received this email from them
        auths = addr.str_to_pairlist(mgr.get(fname, mailfile.FROM_L))
        for auth in auths:
          if auth in  addrpr_to_counts:
            addrpr_to_counts[auth][1] += 1
          else:
            addrpr_to_counts[auth] = [0,1]
      

    if do_threads or do_ids:
      msgid = mgr.get(fname, mailfile.MSG_ID_L)
      ids_to_fnames[msgid] = fname
      if do_threads:
        for i_ref in mgr.get_references(fname):
          if i_ref in ids_to_refbysets:
            ids_to_refbysets[i_ref].add(msgid)
          else:
            ids_to_refbysets[i_ref] = set([msgid]) # start with one elem

    if do_lists:
      state = mgr.get(fname, mailfile.STATE_L)
      if state == "open":
        openlist.append(fname)
      elif state.startswith("scheduled"):
        datestr = state.split()[1]
        scheduledlist.append((fname,datestr))
    
    # do attachments
    if olddir is not None and newdir is not None:
      alist = mgr.get_attachlist(fname)
      if len(alist) > 0:
        blist = [replace_prefix(s, olddir, newdir, fname) for s in alist]
        mgr.set_attachlist(fname, blist)

    # release memory every so often
    msgs_in_mem += 1
    if msgs_in_mem >= MAX_MSGS_IN_MEM:
      mgr = mailfile.MailMgr()
      msgs_in_mem = 0
  # end loop

  if do_tags:
    tagmgr = TagMgr()
    tagmgr.reset_tags_to(list(used_tagset))

  if do_book:
    addr.write_addr_file(addrpr_to_counts)

  if do_threads:
    update_refby_info(ids_to_fnames, ids_to_refbysets)

  if do_ids:
    msgids.rewrite_from_dict(ids_to_fnames)
    
  if do_lists:
    chstate.rewrite_open_list(openlist)
    chstate.rewrite_scheduled_list(scheduledlist)


def migrate(newdir):
  olddir = pathcfg.bomail_data_base
  if not os.path.exists(newdir):
    print("Error: path does not exist: " + newdir)
  print("Moving data files...")
  try:
    newpath = os.path.join(newdir, "bomail")
    shutil.move(pathcfg.bomail_data_base, newpath)
    print("done!")
  except:
    print("Unable to move data directory!")
    traceback.print_exc()
    exit(0)
  bomail.config.config.load_paths(newdir)
  try:
    edit_config.change_option("data_location", newpath)
    print("successfully edited config file")
  except:
    print("Moved data, but unable to edit configuration file bomail.conf!")
    print("Please make change in " + pathcfg.config_file)
    print("data_location = " + newpath)
    print()
    traceback.print_exc()

  # now our lists and msgid file are out of date
  print("updating metadata...")
  update(False, False, False, True, True, olddir=olddir, newdir=newdir)
  print("done!")
  


def main_cli():
  if len(sys.argv) <= 1 or "-h" in sys.argv:
    print(usage_str)
    exit(0)

  if "-l" in sys.argv:
    if os.path.exists(pathcfg.error_log_file):
      os.remove(pathcfg.error_log_file)
    if os.path.exists(pathcfg.acts_log_file):
      os.remove(pathcfg.acts_log_file)
    print("deleted log files")
    exit(0)

  do_tags, do_book, do_ids, do_threads, do_lists = False, False, False, False, False
  if "-t" in sys.argv or "-a" in sys.argv:
    do_tags = True
  if "-b" in sys.argv or "-a" in sys.argv:
    do_book = True
  if "-d" in sys.argv or "-a" in sys.argv:
    do_threads = True
  if "-i" in sys.argv or "-a" in sys.argv:
    do_ids = True
  if "-s" in sys.argv or "-a" in sys.argv:
    do_lists = True

  if any([do_tags, do_book, do_threads, do_ids, do_lists]):
    print("updating metadata...")
    update(do_tags, do_book, do_threads, do_ids, do_lists)
    print("done!")
 
  if "-m" in sys.argv:
    i = sys.argv.index("-m")
    if i+1 >= len(sys.argv):
      print("Missing path argument (use -h for help)")
      exit(0)
    migrate(sys.argv[i+1])

  
if __name__ == "__main__":
  main_cli()




####
# bomail.guistuff.viewinterface
#
# Interface for a view of a tab or thread, inside the gui.
####


class ViewInterface:
  # get list of currently marked filenames (or filename under cursor)
  # for operations to apply to
  def get_curr_filenames(self):
    raise Exception("Not implemented")

  # return mode, note, new_view
  # mode: string (what to redraw, 'all', 'tab', 'note')
  # note: string (what note to redraw)
  # new_view: a ViewInterface object such as self
  def process_key(self, key):
    raise Exception("Not implemented")

  # note the messages must be reloaded
  def mark_obsolete(self):
    raise Exception("Not implemented")

  # called when new messages are received
  # need previous data??
  # no return variables
  def update_for_new(self, filelist):
    raise Exception("Not implemented")

  # called when messages/metadata are changed
  # need previous data??
  # no return variables
  def update_for_change(self, filelist):
    raise Exception("Not implemented")

  # called when messages are deleted from system
  # need previous data??
  # return True if our view is now empty, False otherwise
  def update_for_trash(self, filelist):
    raise Exception("Not implemented")

  def redraw(self, gui):
    raise Exception("Not implemented")

  # return the modified note
  def mod_note(self, note):
    raise Exception("Not implemented")








#--------------------------------------------------------------
bomail_tabcomplete_str = """

_bomail()
{
    local cur=${COMP_WORDS[COMP_CWORD]}
    COMPREPLY=( $(compgen -W "gui process check_sched search chstate compose mailfile send stats meta" -- $cur) )
}
complete -F _bomail bomail

"""


#--------------------------------------------------------------
sample_mailhandlers_str = """## bomail mail-handling configuration for processing of new email.
## Lines starting with '#' are comments.
## To 'uncomment' a line, delete the initial #.

## ============================================================
## Global options - comment to turn off.

## Tag replies to a thread with parent message's tags
-autotagreplies

## Use tags from last line of email if of the form "tags: tag1, ..."
-lastlinetags

## Automatically close any sent mail
-close-sent

## ============================================================
## Custom handlers - you write these.
##
## Each handler has blank lines before and after it, none in the middle.
##
## It has one or more search string lines (see bomail search -h),
## followed by (optional) a state change which can be
## "open", "close", or "schedule ..." (see bomail help datestr),
## followed by (optional) "tag first-tag, second-tag, ...".
## Use OR with multiple search lines. Examples:
##
## -from Facebook
## tag social
##
## -from Twitter
## OR -subject Twitter
## close
##
## -from Instagram
## OR -subject Instagram
## OR -to instagram@domain.com
## schedule p1d
## tag social, photos
"""

#--------------------------------------------------------------
sample_config_str = """## bomail configuration file
##
## All lines beginning with # are comments and are ignored.
## To uncomment a line, delete the initial #.



## ============================================================
## User info: must edit this section!

name = YOUR NAME
email_addr = USER@DOMAIN.COM

## Send email from this server.
## If smtp_servername = localhost, the other options are ignored.
## For SMTP, read username/pass from file, or enter them here.
smtp_servername = smtp.gmail.com
smtp_port = 587
smtp_userpass_file = HOMEDIR/.getmail/getmailrc
#smtp_username = USER or USER@DOMAIN.COM
#smtp_password = PASSWORD

#alias_addresses = ALSO_ME@DOMAIN.COM, ALSO_METOO@DOMAIN.COM

## Additional aliases here. Add numbers 2,3,...
## Just edit "From" line in email to send from these instead.
#name2 = YOUR NAME
#email_addr2 = ALTERNATE@DOMAIN.COM
#smtp_servername2 = smtp.office365.com
#smtp_port2 = 993
#smtp_userpass_file2 = HOMEDIR/.getmail/getmailrc_EDITME
#smtp_username2 = USER2 or USER2@DOMAIN.COM
#smtp_password2 = PASSWORD2



## ============================================================
## Organization options: should confirm these are okay

data_location = DATADIR
new_rawmail_location = NEWMAILDIR
processed_rawmail_location = OLDMAILDIR



## ============================================================
## Important UI options

## Comment next line to turn off threading (conversations)
threads_on

read_program = less
edit_program = vim

## Uncomment next line to use vim-style navigation (hjkl for left/down/up/right)
#hjkl_navigation

## Options are light1, light2, dark1, dark2
colorscheme = dark1



## ============================================================
## Other UI options

## Options are none, some, all
strip_newlines_in_msg_preview = some

## Num lines (including subject/author/date and preview lines) per message
total_lines_per_msg = 5
threadview_total_lines_per_msg = 10

## Number of characters to use ("len") and for spacing ("pad")
date_len = 20
date_pad = 2
author_len = 24
author_pad = 2

## Comment to turn off
#hline_between_msgs
skipline_between_msgs
show_tags
tags_on_topline


"""






# bomail.config.optconfig
# Options not in bomailrc (may edit)


# how many new emails to process at a time, -1 for no limit
process_new_limit = -1

# how many new emails to process in a batch, decrease to save memory
# for very large imports
process_batch_size = 1000

# tag drafts with same tags as parent
autotag_draft_replies = True

## Replying to emails
## Will replace %from, %to, %cc, %date, %time, %subject, %body
##   with the corresponding text from the quoted email.

# quote_format = """\n
# %from wrote on %date at %time:
# %body"""

quote_format = """\n
From: %from
Date: %date at %time
Subject: %subject

%body"""



## %a is day of week; %Y,%m,%d are year/month/date
## %H,%M are hour/minute; %Z is timezone name, %z is timezone offset +hh
quote_date_fmt = "%a, %Y-%m-%d"
quote_time_fmt = "%H:%M UTC%z"
quote_line_prefix = "> "





import os, sys

from bomail.config.config import pathcfg
from bomail.cli import gui,process,check_sched,search,chstate,mailfile,compose,send,meta,stats
import bomail.util.datestuff as datestuff


usage_str = """
Usage:
Use bomail <command> -h for help on that command.

Commands:

  bomail gui          Launch interface (if in doubt, run this)

  bomail process      Check for and process new emails
  bomail check_sched  Check 'scheduled' emails and unschedule them
  bomail search       Search
  bomail chstate      Change state of emails (open/closed/scheduled/trash)
  bomail compose      Compose new emails as blank, reply, forward
  bomail mailfile     See fields of an email or add/remove tags
  bomail send         Send drafts
  bomail meta         Change data location, cleanup metadata
  bomail stats        See some stats about email

  bomail help datestr Print help about using relative dates
  bomail help tags    Print help about 'dirlike' tags
  bomail help meta    Print help about mail filters and other maintenence

Configuration: """ + pathcfg.config_file + """
Data location: """ + pathcfg.bomail_data_base + """
Email filters: """ + pathcfg.handlers_file + """

"""

taghelp_str = """
Using tags:

An email can have unlimited tags.
Searching for emails with a given tag must match
exactly (so "bomail search -t happy" does not match
emails tagged happydays, but does match emails tagged happy.)

Twist: 'directory-like' tags.
Whenever your tag has a / character, it matches searches
for tags of a prefix up to a /.

Example: anything tagged social/twitter, social/facebook,
or social/mastodon will match "bomail search -t social".
However, none of those will match "bomail search -t facebook".
(Use "bomail search -t social/facebook" instead.)

"""

metahelp_str = """
Metadata and maintenence info:

-- Mail filters/handling --
Please modify the mail-handlers file located in
""" + pathcfg.handlers_file + """

-- Trash --
When trashed, *the original MIME email file is not touched.*
The mailfile and stripped attachments *are not permanently deleted.*
They are placed in """ + pathcfg.trash_dir + """
which you may clear at your convenience.

-- Metadata files --
These are in """ + pathcfg.metadata_dir + """
To refresh metadata, run bomail meta -h
  mail-handlers.txt  Handling new mail - please modify!
  error_log.txt      Errors are logged here. May modify/delete.
  acts_log.txt       Logs all actions, can grow large. May modify/delete.
  addr_book.txt      Used to suggest recipients. No harm to modifying.
  tags.txt           Used to suggest tags. No harm to modifying.
  msg_ids.txt        Do not modify (maps msg id to filename location)
  openlist.txt       Do not modify (list of "open" state emails, for speed)
  scheduledlist.txt  Do not modify ("scheduled" state emails and time to open)
"""

def main():
  if len(sys.argv) <= 1:
    print(usage_str)
    exit(0)

  if sys.argv[1] == "help":
    if len(sys.argv) <= 2:
      print(usage_str)
    elif sys.argv[2] == "datestr":
      print(datestuff.datestr_str)
    elif sys.argv[2] == "tags":
      print(taghelp_str)
    elif sys.argv[2] == "meta":
      print(metahelp_str)
    else:
      print(usage_str)
    exit(0)

  # hacky but simple: make programs act like they were
  # invoked from command line
  sys.argv = sys.argv[1:]

  if sys.argv[0] == "gui":
    gui.main_cli()
  elif sys.argv[0] == "process":
    process.main_cli()
  elif sys.argv[0] == "check_sched":
    check_sched.main_cli()
  elif sys.argv[0] == "search":
    search.main_cli()
  elif sys.argv[0] == "chstate":
    chstate.main_cli()
  elif sys.argv[0] == "mailfile":
    mailfile.main_cli()
  elif sys.argv[0] == "compose":
    compose.main_cli()
  elif sys.argv[0] == "meta":
    meta.main_cli()
  elif sys.argv[0] == "send":
    send.main_cli()
  elif sys.argv[0] == "stats":
    stats.main_cli()

  else:
    print(usage_str)

if __name__ == "__main__":
  main()




####
# bomail.guistuff.display
#
# Displaying various stuff on screen.
####

import sys
import time
import curses
import curses.textpad
import subprocess
import runpy
import tempfile
import os
import re
import time
import textwrap, time

from bomail.config.config import guicfg
import bomail.cli.chstate as chstate
import bomail.cli.mailfile as mailfile
import bomail.cli.compose as compose
import bomail.cli.search as search
import bomail.util.addr as addr
import bomail.util.tags as tags
import bomail.util.search_opts as search_opts

def rpad(s,n):
  return s if len(s) >= n else s + " "*(n-len(s))

def lpad(s,n):
  return s if len(s) >= n else " "*(n-len(s)) + s

def pad(key_instr_pr, nkey, ninstr):
  keys,instr = key_instr_pr
  return lpad(keys,nkey) + " " + rpad(instr,ninstr)

ldur_chars = guicfg.LEFT_KEY + guicfg.DOWN_KEY + guicfg.UP_KEY + guicfg.RIGHT_KEY
way_ldur_chars = guicfg.WAY_LEFT_KEY + guicfg.WAY_DOWN_KEY + guicfg.WAY_UP_KEY + guicfg.WAY_RIGHT_KEY
page_options = "expand, fast move"
if not guicfg.hjkl_nav:
  ldur_chars = "↓↑→" 
  way_ldur_chars = "⇧ ↔"  # shift+arrows
  page_options = "expand thread"

ldur_k  = ldur_chars, "left/down/up/right"
page_k  = way_ldur_chars, page_options
quit_k  = guicfg.QUIT_KEY, "quit"
tab_k1  = guicfg.TAB_LEFT_KEY + guicfg.TAB_RIGHT_KEY, "go tab left/right"
tab_k2  = guicfg.MOVE_TAB_LEFT_KEY + guicfg.MOVE_TAB_RIGHT_KEY, "push tab left/right"
tab_k3  = guicfg.ADD_TAB_KEY + guicfg.REMOVE_TAB_KEY, "add/remove tab"
tab_k4  = guicfg.EDIT_TAB_KEY + guicfg.SEARCH_IN_TAB_KEY, "edit/search tab"
state_k = guicfg.OPEN_KEY + guicfg.CLOSE_KEY + guicfg.SCHEDULE_KEY + guicfg.TRASH_KEY, "open/close/sched/trash"
tags_k  = guicfg.ADD_TAGS_KEY + guicfg.REMOVE_TAGS_KEY, "add/delete tags"
undo_k  = guicfg.UNDO_KEY + guicfg.REDO_KEY, "undo/redo"
get_k   = guicfg.GET_KEY, "get update"
write_k = guicfg.WRITE_KEY, "write"
send_k  = guicfg.SEND_KEY, "send"
view_k  = guicfg.VIEW_KEY, "view"
mark_k  = guicfg.MARK_KEY + guicfg.MARK_ALL_KEY, "mark/all"
TAB_MSG_INSTR_STR = [
  pad(quit_k, 2,12) + pad(ldur_k, 4,23) + pad(get_k,  2,12) + pad(tab_k1,2,0),
  pad(mark_k, 2,12) + pad(page_k, 4,23) + pad(view_k, 2,12) + pad(tab_k2,2,0),
  pad(undo_k, 2,12) + pad(tags_k, 4,23) + pad(write_k,2,12) + pad(tab_k3,2,0),
  pad(("",""),2,12) + pad(state_k,4,23) + pad(send_k, 2,12) + pad(tab_k4,2,0),
]


TOPLINES = 3
BOTTOMLINES = 7


#--------------------------------------------------------------
# drawing to screen

# insert the given string to the given window with the given attribute
def my_insstr(window, y, x, s, attr=0):
  height, width = window.getmaxyx()
  if y >= height or x >= width:
    return
  window.insstr(y, x, s.encode("utf-8")[:width-x], attr)

#def my_addstr(window, y, x, s, attr):
#  height, width = window.getmaxyx()
#  if y >= height:
#    return
#  window.addstr(y, x, s.encode("utf-8")[:width-x], attr)


# write the array of msg_lines to screen,
# then set all the attributes for each (y, x, width, attr) in attr_data
def write_lines(window, y_min, y_max, y_start, x_start, msg_lines, attr_data):
  h,w = window.getmaxyx()
  for j,line in enumerate(msg_lines):
    y = y_start + j
    if y >= y_min and y < y_max:
      try:
        my_insstr(window, y, x_start, line.encode("utf-8"), 0)
      except:
        my_insstr(window, y, x_start, line, 0)
  for j,x,width,attr in attr_data:
    x = x_start + x
    y = y_start + j
    if y >= y_min and y < y_max and x+width < w:
      window.chgat(y, x, width, attr)


def draw_loading_screen(gui):
  height, width = gui.screen.getmaxyx()
  gui.screen.clear()
  gui.screen.refresh()  # this needs to be here, no idea why
  toptabs = "[1]"
  my_insstr(gui.screen, 0, 0, toptabs, gui.attr)
  my_insstr(gui.screen, 2, 0, "-"*curses.COLS, gui.attr)
  gui.screen.chgat(0, 0, 3, curses.A_STANDOUT | gui.attr)
  gui.screen.hline(height-BOTTOMLINES, 0, "-", width, gui.attr)
  gui.screen.hline(height-BOTTOMLINES+2, 0, "-", width, gui.attr)
  redraw_note(gui, "Loading (please wait)...")
  

# mode is all, view, or note
# curr_view implements viewinterface.ViewInterface
def redraw(gui, curr_view, mode, note=""):
  # only redraw overlay if all
  if mode == "all":
    redraw_overlay(gui)
  # only redraw view if all or view
  if mode in ["all", "view"]:
    redraw_view(gui, curr_view)
  # always draw the note
  height,width = gui.note_area.getmaxyx()
  note = curr_view.mod_note(note, width)
  redraw_note(gui, note)  # refreshes screen


# top lines
def redraw_overlay(gui):
  height, width = gui.screen.getmaxyx()
  gui.screen.clear()
  gui.screen.refresh()  # this needs to be here, no idea why
  toptabs = "[1]  [2]  [3]  [4]  [5]  [6]  [7]  [8]  [9]  [0]"
  if len(gui.tabs) <= 10:
    len_tt = len(gui.tabs)*5 - 2
    toptabs = toptabs[:len_tt]
  else:
    toptabs = toptabs + "  [ ]" * (len(gui.tabs) - 10)
  my_insstr(gui.screen, 0, 0, toptabs, gui.attr)
  my_insstr(gui.screen, 1, 0, gui.tabs[gui.tab_ind].search_str, gui.attr)
  my_insstr(gui.screen, 2, 0, "-"*curses.COLS, gui.attr)
  if 5*gui.tab_ind + 3 <= width:
    gui.screen.chgat(0, 5*gui.tab_ind, 3, curses.A_STANDOUT | gui.attr)  # highlight tab number
  gui.screen.hline(height-BOTTOMLINES, 0, "-", width, gui.attr)
  gui.screen.hline(height-BOTTOMLINES+2, 0, "-", width, gui.attr)


# tab section and commands
def redraw_view(gui, curr_view):
  gui.tab_area.clear()
  gui.tab_area.refresh()
  curr_view.redraw(gui)
  draw_bottom_text(gui, TAB_MSG_INSTR_STR)


def redraw_tab_lin(thread_view, gui):
  thread_view.ensure_display_loaded()
  height, screen_width = gui.tab_area.getmaxyx()

  # can assume we have > 0 files
  subj_str = gui.mail_mgr.get(thread_view.repr_file, mailfile.SUBJ_L)[:screen_width]
  my_insstr(gui.tab_area, 0, 0, subj_str, curses.A_BOLD | guicfg.DEFAULT_CLR_PAIR)
  gui.tab_area.hline(1, 0, "-", screen_width)
  offset = 2

  curr_y = thread_view.top_y_coord + offset
  for ind,tup in enumerate(thread_view.file_data):
    if tup[2]:  # is collapsed
      msg_lines, orig_attr_data = tup[3]  # collapsed disp_data
    else:
      msg_lines, orig_attr_data = tup[4]  # expanded disp data
    attr_data = list(orig_attr_data)
    if tup[0] in gui.unread_set:  # is_unread
      attr_data = [(y, x, w, a | curses.A_BOLD if y==0 else a) for y,x,w,a in attr_data]
    if ind == thread_view.cursor_ind:
      attr_data = [(y,x,screen_width,(attr if y>0 else attr|guicfg.cursor_attr)) for y,x,screen_width,attr in attr_data]
    write_lines(gui.tab_area, offset, height, curr_y, 0, msg_lines, attr_data)
    if curr_y >= 0 and curr_y < height and tup[1]:  # is marked
      attr = curses.A_BOLD | gui.attr
      if ind == thread_view.cursor_ind:
        attr |= guicfg.cursor_attr
      gui.tab_area.addch(curr_y, screen_width - 2, "X", attr)
    if curr_y-1 >= 0 and curr_y+1 < height and tup[2]:  # collapsed
      gui.tab_area.addstr(curr_y+1, 1, "<--")

    curr_y += len(msg_lines)

  gui.tab_area.refresh()
  draw_bottom_text(gui, TAB_MSG_INSTR_STR)


# TODO - implement!
def redraw_tab_tree(thread_view, gui):
  return redraw_tab_lin(thread_view, gui)


def draw_bottom_text(gui, lines):
  height, width = gui.commands_area.getmaxyx()
  gui.commands_area.clear()
  for i,line in enumerate(lines):
    if i >= height:
      break
    my_insstr(gui.commands_area, i, 0, lines[i], gui.attr)
  gui.commands_area.refresh()


# add "message x of y" to end of note
# return new note
def mod_note(gui, note):
  mytab = gui.tabs[gui.tab_ind]
  if guicfg.threads_on:
    if mytab.mode == "thread list":
      num_msgs = len(mytab.file_data)
      ind = mytab.cursor_ind
      count_note = "No threads" if num_msgs == 0 else "Thread " + str(ind+1) + " of " + str(num_msgs)
    elif mytab.mode == "one thread":
      num_msgs = len(mytab.file_data[mytab.cursor_ind][0].all_files)
      ind = mytab.file_data[mytab.cursor_ind][0].cursor_ind
      count_note = "No messages" if num_msgs == 0 else "Message " + str(ind+1) + " of " + str(num_msgs)
  else:
    num_msgs = len(mytab.file_data)
    ind = mytab.cursor_ind
    count_note = "No messages" if num_msgs == 0 else "Message " + str(ind+1) + " of " + str(num_msgs)
  height, width = gui.screen.getmaxyx()
  gap = width - (len(note) + len(count_note) + 4)
  if gap >= 0:
    note += " "*(gap + 3) + count_note
  return note



def redraw_note(gui, note):
  gui.note_area.clear()
  my_insstr(gui.note_area, 0, 0, note, gui.attr)
  gui.note_area.refresh()
  gui.screen.refresh()




####
# bomail.cli.chstate
#
# Change state of files (open/closed/scheduled/trash).
####

import sys
import shutil
import os, subprocess
import tempfile
from subprocess import Popen, PIPE

from bomail.config.config import pathcfg
import bomail.cli.mailfile as mailfile
import bomail.util.remove_lines as remove_lines
import bomail.util.merge_lines as merge_lines
import bomail.util.util as util
import bomail.util.datestuff as datestuff

usage_str = """
Change the state of an email to open, closed, or scheduled.
Reads a list of filenames from stdin, one per line.
Call with no args or -h to print this help.

  bomail chstate -o           # make filenames open
  bomail chstate -c           # make filenames closed
  bomail chstate -t           # trash filenames
  bomail chstate -ut          # untrash filenames
  bomail chstate -s datestr   # schedule filename for datestr (e.g. yyyy-mm-dd)
"""


# remove files from open and scheduled lists
def remove_from_lists(flist):
  if os.path.exists(pathcfg.openlist_file):
    remove_lines.do_sorted(pathcfg.openlist_file, sorted(flist))
  if os.path.exists(pathcfg.scheduledlist_file):
    remove_lines.do_sched(pathcfg.scheduledlist_file, set(flist))

def check_flist(flist):
  result = []
  for fname in flist:
    if os.path.exists(fname):
      result.append(fname)
    else:
      util.err_log("Nonexistent filename: " + fname)
  return result

# change all files' state to open
def make_open(flist, mgr, remove_old=True):
  flist = check_flist(flist)
  if remove_old:
    remove_from_lists(flist)
  # sort openlist by lexicographic
  merge_lines.do(pathcfg.openlist_file, sorted(flist))
  for f in flist:
    mgr.set(f, mailfile.STATE_L, "open")


# change all files' state to closed
def make_closed(flist, mgr):
  flist = check_flist(flist)
  remove_from_lists(flist)
  for f in flist:
    mgr.set(f, mailfile.STATE_L, "closed")


def scheduled_list_str(fname, sched_datestr):
  return sched_datestr + " " + fname


# schedule all files for dateobj, a datetime object
def schedule(flist, dateobj, mgr):
  flist = check_flist(flist)
  remove_from_lists(flist)
  datestr = datestuff.get_printed_schedulestr(dateobj)
  # sort scheduledlist
  lines = [scheduled_list_str(f, datestr) for f in flist]
  merge_lines.do(pathcfg.scheduledlist_file, lines)
  for f in flist:
    mgr.set(f, mailfile.STATE_L, "scheduled " + datestr)

# given (unsorted) list of open files, rewrite our save file
def rewrite_open_list(flist):
  flist.sort()  # ascending sort
  with open(pathcfg.openlist_file, "w") as f:
    f.write("\n".join(flist))

# given (unsorted) list of pairs (fname, datestr), rewrite save
def rewrite_scheduled_list(prlist):
  slist = [scheduled_list_str(pr[0], pr[1]) for pr in prlist]
  slist.sort(reverse=True)  # newest first
  with open(pathcfg.scheduledlist_file, "w") as f:
    f.write("\n".join(slist))

# trash all files, including remove them from msg_ids_file
# return corresponding list of new full filenames (in trash folder)
def trash(flist, mgr):
  remove_from_lists(flist)
  os.makedirs(pathcfg.trash_dir, exist_ok=True)
  newnames = []
  idlist = [mgr.get(fname, mailfile.MSG_ID_L) for fname in flist]
  for f in flist:
    # 1. move/write the file
    d = mgr.get_all(f)
    datestr = d[mailfile.DATE_L]
    msg_id = d[mailfile.MSG_ID_L]
    is_draft = f.endswith(mailfile.DRAFT_EXT)
    new_fullname = mailfile.get_trashfilename(datestr, msg_id, is_draft)
    new_dir = os.path.dirname(new_fullname)
    os.makedirs(new_dir, exist_ok=True)
    if os.path.exists(f):
      util.mv_file(f, new_fullname)
    else:
      mailfile.write_to_file(new_fullname, d)
    # 2. move attachments
    if len(d[mailfile.ATTACH_L].strip()) > 0:
      attach_dir = mailfile.get_attach_dir(datestr, msg_id)
      if os.path.exists(attach_dir):
        trash_attach_dir = mailfile.get_trash_attach_dir(datestr, msg_id)
        os.makedirs(os.path.dirname(trash_attach_path[:-1]), exist_ok=True)  # make sure to get the parent
        util.mv_dir(attach_dir, trash_attach_dir)  # make the destination
    newnames.append(new_fullname)
  mgr.update_for_trashed(flist, idlist)
  return newnames


# given list of filenames
def do_untrash(trashed_flist, mgr):
  datalist = [mailfile.read_file(f) for f in trashed_flist]
  create_flist = [mailfile.get_filename(d[mailfile.DATE_L], d[mailfile.MSG_ID_L], f.endswith(mailfile.DRAFT_EXT)) for f,d in zip(trashed_flist, datalist)]
  untrash(create_flist, datalist, mgr)


# given list of new relative filenames and data to put in them
def untrash(create_flist, datalist, mgr):
  # 1) create mailfiles
  create(create_flist, datalist, mgr)
  # 2) move attachments back if needed
  for fname,d in zip(create_flist,datalist):
    if len(d[mailfile.ATTACH_L].strip()) > 0:
      is_draft = fname.endswith(mailfile.DRAFT_EXT)
      datestr = d[mailfile.DATE_L]
      msg_id = d[mailfile.MSG_ID_L]
      trash_attach_dir = mailfile.get_trash_attach_dir(datestr, msg_id)
      attach_dir = mailfile.get_attach_dir(datestr, msg_id)
      for attach_f in os.listdir(trash_attach_dir):
        util.mv_file(os.path.join(trash_attach_dir,attach_f),
                     os.path.join(attach_dir,attach_f))

# the opposite of trash basically
# given a list of filenames and the data
# do error checking, call mgr.create(), and set state open
def create(flist, datalist, mgr):
  for fname,d in zip(flist, datalist):
    if os.path.exists(fname):
      s = ["Creating file, already exists: " + fname]
      d2 = mailfile.read_file(fname)
      for i in range(len(d)):
        if d2[i] != d[i]:
          s.append(">>>> old version " + mailfile.fields[i])
          s.append(d2[i])
          s.append(">>>> new version:")
          s.append(d[i])
      util.err_log("\n".join(s))
    elif mgr.ids.has(d[mailfile.MSG_ID_L]):
      util.err_log("Creating file, non-identical file with same message ID already exists: " + mgr.ids.get(d[mailfile.MSG_ID_L]) + "\n"
                   + "New data:\n  " + "\n  ".join(d) + "\n")
    else:
      mgr.create(flist, datalist)
  make_open(flist, mgr, False)  # False == don't try to remove old ones from list
 

def main(args, flist, mgr):
  flist.sort()
  if args[0] == "-o":
    make_open(flist, mgr)
  elif args[0] == "-c":
    make_closed(flist, mgr)
  elif args[0] == "-t":
    trash(flist, mgr)
  elif args[0] == "-ut":
    do_untrash(flist, mgr)  # filelist are all currently in trash
  elif args[0] == "-s":
    sched = datestuff.parse_schedstr_to_utcobj(args[1])
    if sched is None:
      print("Could not parse scheduling date string!")
      exit(0)
    schedule(flist, sched, mgr)


def main_cli():
  if len(sys.argv) < 2 or "-h" in sys.argv:
    print(usage_str)
    exit(0)
  mgr = mailfile.MailMgr()
  # remove each trailing newline
  main(sys.argv[1:], [s[:-1] for s in sys.stdin.readlines()], mgr)

if __name__ == "__main__":
  main_cli()




####
# bomail.cli.gui
#
# Launch the user interface (which is text based, so not a GUI. I know)
####

import sys
import os
import locale
import curses

import bomail.guistuff.mygui as mygui

gui_helpstr = """
Quitting
--------
q       quit the program. prints the filename(s) under the cursor

Undo and redo
-------------
You can undo/redo most actions like adding/removing tabs,
tagging messages, deleting messages. History is infinite.
u       undo last action (if possible)
r       redo action that was just undone

Tabs and searching
-----------------
Tabs are numbered across the top. Each tab has a search query and
the tab lists all matching messages (or threads).
1..9, 0 switch to that tab number
[ ]     switch one tab to the left and right
{ }     "push" this tab left or right (rearranging tabs)
+       add a new tab. will prompt for the search query
-       delete current tab
e       edit the current tab's search query
/       search within tab. Prompts for a string, then
        creates a new tab with the search results.

Navigation - messages/threads
-----------------------------
v       view thread, or open viewing program to view message.
Arrow keys scroll up/down.
If threading (conversations) is on, then right/left go into/out of
the view of a single thread. In that view, right/left will expand/
collapse individual messages. Shift+right and Shift+left will expand/
collapse all messages.

In ~/.bomailrc you can turn on hjkl_navigation (vim-style) instead
of arrow keys. In this case J and K scroll up/down fast, or
line-by-line in thread view.

Sending/receiving mail
----------------------
Note bomail does not retrieve any mail from the server, it just
looks in the new mail folder. To fetch mail, use e.g. getmail.
g       get updates: look for new mail, open scheduled messages
S       (capital S) send message (if draft); prompts you to confirm

Writing/modifying drafts
------------------------
w       "write" key. Either compose new draft or edit existing.
--If current message is not a draft, prompts these options:
n       compose a brand-new draft unrelated to this message
a       reply-all
o       reply-one (only reply to the sender)
r       reply-recipients (reply to all but sender; for mailing lists)
f       forward
[any other key cancels]
--If current message is a draft, prompts these options:
n       compose a brand-new draft unrelated to this message
e       edit this draft (calls the program chosen in ~/.bomailrc)
        note you can hand-edit the draft instead of the following...
s       edit subject line
a       add attachment(s)
x       remove attachment(s)
t       add new TO: recipient(s)
c       add new CC: recipient(s)
b       add new BCC: recipient(s)
y       remove TO: recipient(s)
v       remove CC: recipient(s)
m       remove BCC: recipient(s)
[any other key cancels]

Message metadata
-------------------------------
You can mark multiple messages and apply the following operations
to all marked messages. If none are marked, the operations apply
to the current message under the cursor.
x       toggle current message/thread as marked/unmarked
X       toggle all messages/threads in tab as marked/unmarked
o       place message(s) in open state (it matches "-open" queries)
c       place in closed state
s       place in scheduled state; prompts for date and time.
        See 'bomail help datestr'
t       move message(s) to trash (located in your bomail/trash folder)
a       add tag(s) to message(s). Opens dialogue to enter the tag
d       delete tag(s) from message(s).

"""

def main(screen):
  return mygui.main(screen)


def main_cli():
  if len(sys.argv) >= 2:
    print(gui_helpstr)
    return
  os.environ.setdefault('ESCDELAY', '25')
  locale.setlocale(locale.LC_ALL, '')
  result = curses.wrapper(main)
  os.system("clear")  # clear console and print filenames or error message
  print()
  print(result)


if __name__ == "__main__":
  main_cli()




####
# bomail.util.msgids
#
# Given a message-id, look up if we have an email file for it.
# Basically a wrapper for reading and writing to plain text file
#
# line format: msgid filename
####

import os, sys

from bomail.config.config import pathcfg
import bomail.util.merge_lines as merge_lines
import bomail.util.remove_lines as remove_lines
import bomail.util.util as util


def msgid_file_line(msgid, filename):
  return msgid.strip() + " " + filename


def get_idfile_lines(msg_id_list, filename_list):
  return sorted([msgid_file_line(m, f) for m,f in zip(msg_id_list,filename_list)])


# given a dictionary mapping all message-ids to their filenames, rewrite the file
def rewrite_from_dict(ids_to_fnames):
  slist = [msgid + " " + fname for msgid,fname in ids_to_fnames.items()]
  slist.sort()
  with open(pathcfg.msg_ids_file, "w") as f:
    f.write("\n".join(slist))


class Ids:
  def __init__(self):
    # map msgid to filename
    self.ids_to_fnames = {}
    if os.path.exists(pathcfg.msg_ids_file):
      with open(pathcfg.msg_ids_file) as f:
        for line in f:
          if " " not in line:
            continue
          ind = line.index(" ")
          if ind+1 >= len(line):
            continue
          msgid = line[:ind].strip()
          fname = line[ind+1:].strip()
          self.ids_to_fnames[msgid] = fname

  def has(self, msg_id):
    return msg_id in self.ids_to_fnames

  def get(self, msg_id):
    return self.ids_to_fnames[msg_id]

  def add(self, msg_id_list, fname_list):
    for m,f in zip(msg_id_list, fname_list):
      self.ids_to_fnames[m] = f
    merge_lines.do(pathcfg.msg_ids_file, get_idfile_lines(msg_id_list, fname_list))

  def remove(self, msg_id_list, fname_list):
    for m,fname in zip(msg_id_list, fname_list):
      if m in self.ids_to_fnames:
        del self.ids_to_fnames[m]
      elif not fname.endswith("draft"):
        # if it's not a draft, then it should have been in our database
        util.err_log("trying to remove a message id that's not in our database:\n" + m)
    remove_lines.do_sorted(pathcfg.msg_ids_file, get_idfile_lines(msg_id_list, fname_list))


